#*******************************************************************************
# Copyright 2014-2020 Intel Corporation
# All Rights Reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License"), the following terms apply:
#
# You may not use this file except in compliance with the License.  You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
#*******************************************************************************

#!/bin/bash

# if dpc++ vars path is specified
if [ ! -z "${DPCPPROOT}" ]; then
    source ${DPCPPROOT}/env/vars.sh
fi

# if DAALROOT is specified
if [ ! -z "${DAALROOT}" ]; then
    conda remove daal --force -y
    source ${DAALROOT}/env/vars.sh
fi

# if TBBROOT is specified
if [ ! -z "${TBBROOT}" ]; then
    conda remove tbb --force -y
    source ${TBBROOT}/env/vars.sh
fi


set -ex



test -f ${PREFIX}/lib/python*/site-packages/onedal/_onedal_py_host*.so
cd tests
python -c "import daal4py"
mpirun -n 4 python -m unittest discover -v -p spmd*.py
python -m unittest discover -v -p 'test*.py'
pytest --pyargs ../daal4py/sklearn/
pytest --pyargs ../onedal/
python ../examples/daal4py/run_examples.py
python -m daal4py ../examples/daal4py/sycl/sklearn_sycl.py
exit 0
